@@ -32,6 +32,27 @@ $tab }
 ";
 }
 
+sub call_subrule_no_capture {
+    my ( $subrule, $tab, @param ) = @_;
+    $subrule = "\$_[4]->" . $subrule unless $subrule =~ / :: | \. | -> /x;
+    $subrule =~ s/\./->/;   # XXX - source filter
+
+    push @param, 1 if @param == 1;  # odd number of elements in hash
+    #print "PARAM: ",Dumper(@param);
+
+    return 
+"$tab sub{ 
+$tab     my \$prior = \$::_V6_PRIOR_;
+$tab     my \$param = { \%{ \$_[7] || {} }, args => {" . 
+            join(", ",@param) . "} };
+$tab     \$_[3] = $subrule( \$_[0], \$param, \$_[3],  );
+$tab     \$_[3]->data->{match} = [];
+$tab     \$_[3]->data->{named} = {};
+$tab     \$::_V6_PRIOR_ = \$prior;
+$tab }
+";
+}
+
 sub emit {
     my ($grammar, $ast) = @_;
     # runtime parameters: $grammar, $string, $state, $arg_list
@@ -41,6 +62,7 @@ sub emit {
     local %capture_seen = ();
     #print "emit capture_to_array $capture_to_array\n";
     # print "emit: ", Dumper($ast);
+    #die emit_rule( $ast, '    ' );
     
     return 
 "do {
@@ -52,11 +74,14 @@ sub emit {
         #"    print \"match args: \",Dumper(\@_);\n" .
         "
     my \$tree;
-    if ( defined \$_[3]{p} ) {
+    if (  defined \$_[3]{p} 
+       && ! \$_[3]{continue}
+       ) {
       \$matcher->( \$_[1], \$_[2], \$tree, \$tree, \$_[0], \$_[3]{p}, \$_[1], \$_[3] );
     }
     else {
-      for my \$pos ( 0 .. length( \$_[1] ) ) {
+      \$_[3]{p} ||= 0;
+      for my \$pos ( \$_[3]{p} .. length( \$_[1] ) ) {
         my \$param = { \%{\$_[3]}, p => \$pos };
         \$matcher->( \$_[1], \$_[2], \$tree, \$tree, \$_[0], \$pos, \$_[1], \$param );
         last if \$tree;
@@ -69,7 +94,16 @@ sub emit {
       \$::_V6_MATCH_ = \$tree;
       \$tree->data->{capture} = \\(\$cap->( \$tree ));
     };
-    if ( \$tree ) { \$::_V6_PRIOR_ = \$rule }
+    if ( \$tree ) { 
+      # \$::_V6_PRIOR_ = \$rule 
+
+      my \$prior = \$::_V6_PRIOR_;
+      \$::_V6_PRIOR_ = sub { 
+        local \$main::_V6_PRIOR_ = \$prior; 
+        \$rule->(\@_);
+      };
+
+    }
     return \$tree;
   }
 }
@@ -136,7 +170,18 @@ sub non_capturing_group {
 sub quant {
     my $term = $_[0]->{'term'};
     my $quantifier = $_[0]->{quant}  || '';
-    my $greedy     = $_[0]->{greedy} || '';   # + ?    
+    my $greedy     = $_[0]->{greedy} || '';   # + ?   
+    
+    if ( ref( $quantifier ) eq 'HASH' ) 
+    {
+        die "quantifier not implemented: " . Dumper( $quantifier );
+
+        #return 
+        #    "$_[1] concat(\n" .
+        #    join( ',', ($rul) x $count ) .
+        #    "$_[1] )\n";
+    }
+       
     my $quant = $quantifier . $greedy;
     my $sub = { 
             '*' =>'greedy_star',     
@@ -166,7 +211,7 @@ sub quant {
 }        
 sub alt {
     my @s;
-    
+    # print "Alt: ", Dumper($_[0]);
     my $count = $capture_count;
     my $max = -1;
     for ( @{$_[0]} ) { 
@@ -191,6 +236,7 @@ sub alt {
            join( ',', @s ) .
            "$_[1] ] )\n";
 }        
+sub alt1 { &alt }
 sub concat {
     my @s;
     for ( @{$_[0]} ) { 
@@ -234,7 +280,8 @@ sub variable {
         }
     }
     
-    $value = join('', eval $name) if $name =~ /^\@/;
+    # ??? $value = join('', eval $name) if $name =~ /^\@/;
+
     if ( $name =~ /^%/ ) {
         # XXX - runtime or compile-time interpolation?
         return "$_[1] hash( \\$name )\n" if $name =~ /::/;
@@ -246,7 +293,40 @@ sub variable {
     return "$_[1] constant( '" . $value . "' )\n";
 }
 sub special_char {
-    my $char = substr($_[0],1);
+    my ($char, $data) = $_[0] =~ /^.(.)(.*)/;
+
+    return  "$_[1] perl5( '\\N{" . join( "}\\N{", split( /\s*;\s*/, $data ) ) . "}' )\n"
+        if $char eq 'c';
+    return  "$_[1] perl5( '(?!\\N{" . join( "}\\N{", split( /\s*;\s*/, $data ) ) . "})\\X' )\n"
+        if $char eq 'C';
+
+    return  "$_[1] perl5( '\\x{$data}' )\n"
+        if $char eq 'x';
+    return  "$_[1] perl5( '(?!\\x{$data})\\X' )\n"
+        if $char eq 'X';
+
+    return special_char( sprintf("\\x%X", oct($data) ) )
+        if $char eq 'o';
+    return special_char( sprintf("\\X%X", oct($data) ) )
+        if $char eq 'O';
+
+    return  "$_[1] perl5( '(?:\\n\\r?|\\r\\n?|\\x85|\\x{2028})' )\n"
+        if $char eq 'n';
+    return  "$_[1] perl5( '(?!\\n\\r?|\\r\\n?|\\x85|\\x{2028})\\X' )\n"
+        if $char eq 'N';
+
+    # XXX - Infinite loop in pugs stdrules.t
+    #return metasyntax( '?_horizontal_ws', $_[1] )
+    return "$_[1] perl5( '[\\x20\\x09]' )\n" 
+        if $char eq 'h';
+    return "$_[1] perl5( '[^\\x20\\x09]' )\n" 
+        if $char eq 'H';
+    #return metasyntax( '?_vertical_ws', $_[1] )
+    return "$_[1] perl5( '[\\x0A\\x0D]' )\n" 
+        if $char eq 'v';
+    return "$_[1] perl5( '[^\\x0A\\x0D]' )\n" 
+        if $char eq 'V';
+
     for ( qw( r n t e f w d s ) ) {
         return "$_[1] perl5( '\\$_' )\n" if $char eq $_;
         return "$_[1] perl5( '[^\\$_]' )\n" if $char eq uc($_);
@@ -362,7 +442,8 @@ sub named_capture {
         "$_[1] )\n";
 }
 sub negate {
-    my $program = $_[0]{rule};
+    my $program = $_[0];
+    #print "Negate: ", Dumper($_[0]);
     return 
         "$_[1] negate( \n" . 
         emit_rule($program, $_[1]) . 
@@ -387,10 +468,21 @@ sub colon {
         if $str eq '^^';
     return "$_[1] at_line_end() \n"
         if $str eq '$$';
+    return metasyntax( '?_wb_left', $_[1] )
+        if $str eq '<<';
+    return metasyntax( '?_wb_right', $_[1] )
+        if $str eq '>>';
     die "'$str' not implemented";
 }
 sub modifier {
-    my $str = $_[0];
+    my $str = $_[0]{modifier};
+    my $rule = $_[0]{rule};
+
+    return "$_[1] ignorecase( \n"
+        . emit_rule( $rule, $_[1] . '  ' )
+        . " )\n" 
+        if $str eq 'ignorecase';
+
     die "modifier '$str' not implemented";
 }
 sub constant {
@@ -449,13 +541,64 @@ sub metasyntax {
         return;
     }
     if ( $prefix =~ /[-+[]/ ) {   # character class 
-        $cmd =~ s/\.\./-/g;
-        if ( $prefix eq '-' ) {
-	       $cmd = '[^' . substr($cmd, 2);
-	    } 
-        elsif ( $prefix eq '+' ) {
-	       $cmd = substr($cmd, 2);
+        #die "SET regex: $cmd\n";
+        
+        $cmd =~ s/\.\./-/g;  # ranges
+        
+        # TODO - \o \O
+
+        if    ( $cmd =~ /^ \+? \[ \\ c \[ (.*) \] \] /x ) {
+            #$cmd = "(?:\\N{" . join( "}|\\N{", split( /\s*;\s*/, $1 ) ) . "})";
+            $cmd = "[\\N{" . join( "}\\N{", split( /\s*;\s*/, $1 ) ) . "}]";
+        }
+        elsif ( $cmd =~ /^ \+? \[ \\ C \[ (.*) \] \] /x ) {
+            #$cmd = "(?!\\N{" . join( "}|\\N{", split( /\s*;\s*/, $1 ) ) . "})\\X";
+            $cmd = "[^\\N{" . join( "}\\N{", split( /\s*;\s*/, $1 ) ) . "}]";
+        }
+        elsif ( $cmd =~ /^ -  \[ \\ C \[ (.*) \] \] /x ) {
+            #$cmd = "(?:\\N{" . join( "}|\\N{", split( /\s*;\s*/, $1 ) ) . "})";
+            $cmd = "[\\N{" . join( "}\\N{", split( /\s*;\s*/, $1 ) ) . "}]";
+        }
+        elsif ( $cmd =~ /^ -  \[ \\ c \[ (.*) \] \] /x ) {
+            #$cmd = "(?!\\N{" . join( "}|\\N{", split( /\s*;\s*/, $1 ) ) . "})\\X";
+            $cmd = "[^\\N{" . join( "}\\N{", split( /\s*;\s*/, $1 ) ) . "}]";
+        }
+
+        
+        elsif ( $cmd =~ /^ \+? \[ \\ x \[ (.*) \] \] /x ) {
+            $cmd = "(?:\\x{$1})";
+        }
+        elsif ( $cmd =~ /^ \+? \[ \\ X \[ (.*) \] \] /x ) {
+            $cmd = "(?!\\x{$1})\\X";
+            #$cmd = "[^\\x{$1}]";
+        }
+        elsif ( $cmd =~ /^ -  \[ \\ X \[ (.*) \] \] /x ) {
+            $cmd = "(?:\\x{$1})";
+            #$cmd = "[\\x{$1}]";
+        }
+        elsif ( $cmd =~ /^ -  \[ \\ x \[ (.*) \] \] /x ) {
+            $cmd = "(?!\\x{$1})\\X";
+        }
+        
+        
+        elsif ( $cmd =~ /^ - \s* \[ (.*) /x ) {
+           $cmd = '[^' . $1;
+        } 
+        elsif ( $cmd =~ /^ - \s* (.*) /x ) {
+           my $name = $1;
+           $cmd = ( $name =~ /^is/ )
+                ? "\\P{$name}"
+                : "[^[:$name:]]";
+        } 
+        elsif ( $cmd =~ /^ \+ \s* \[ (.*) /x ) {
+           $cmd = '[' . $1;
 	    }
+        elsif ( $cmd =~ /^ \+ \s* (.*) /x ) {
+           my $name = $1;
+           $cmd = ( $name =~ /^is/ )
+                ? "\\p{$name}"
+                : "[[:$name:]]";
+        } 
 	    # XXX <[^a]> means [\^a] instead of [^a] in perl5re
         return "$_[1] perl5( q!$cmd! )\n" unless $cmd =~ /!/;
         return "$_[1] perl5( q($cmd) )\n"; # XXX if $cmd eq '!)'
@@ -466,7 +609,7 @@ sub metasyntax {
             warn "code assertion not implemented";
             return;
         }
-        return call_subrule( $cmd, $_[1] );
+        return call_subrule_no_capture( $cmd, $_[1] );
     }
     if ( $prefix =~ /[_[:alnum:]]/ ) {
         if ( $cmd eq 'cut' ) {
