@@ -36,6 +36,8 @@ sub alternation {
 sub concat {
     my $nodes = shift;
     $nodes = [ $nodes, @_ ] unless ref($nodes) eq 'ARRAY';  # backwards compat
+    return null()      if ! @$nodes;
+    return $nodes->[0] if @$nodes == 1;
     if ( @$nodes > 2 ) {
         return concat(
             concat( [ $nodes->[0], $nodes->[1] ] ),
@@ -46,6 +48,12 @@ sub concat {
         my @state = $_[1] ? @{$_[1]} : ( undef, undef );
         #print "enter state ",Dumper(\@state);
         my $m2;
+        my $redo_count = 0;  
+            # XXX - workaround for t/regex/from_perl6_rules/capture.t test #38:
+            # regex single { o | k | e };
+            # # ...
+            # ok(!( "bokeper" ~~ m/(<?single>) ($0)/ ), 'Failed positional backref');
+
         do {
 
             my %param1 = defined $_[7] ? %{$_[7]} : ();
@@ -54,12 +62,14 @@ sub concat {
             $nodes->[0]->( $_[0], $state[0], @_[2..7] );
             return if ! $_[3] 
                    || $_[3]->data->{abort};
-            my $is_empty = $_[3]->from == $_[3]->to;
 
-            if ( $param1{was_empty} && $is_empty ) {
-                # perl5 perlre says "the following match after a zero-length match
+            my $is_empty = ( $_[3]->from == $_[3]->to );
+            #    && ( $param1{was_empty} )
+            #    ; # fix a problem with '^'
+            if ( $is_empty && $param1{was_empty} ) {
+                #    # perl5 perlre says "the following match after a zero-length match
                 #   is prohibited to have a length of zero"
-                return;
+                return unless $_[3]->from == 0;
             }
 
             my $param = { ( defined $_[7] ? %{$_[7]} : () ), 
@@ -72,14 +82,18 @@ sub concat {
             #print "concat 2: "," \n";
             $nodes->[1]->( $_[0], $state[1], $_[2], $m2, 
                            $_[4], $_[3]->to, $_[6], $param );
+            
+              #return if $is_empty && $m2->from == $m2->to; 
             $state[1] = $m2->state;
             $state[0] = $next_state unless $state[1];
             #print "concat 3: "," \n";
             #print "return state ",Dumper(\@state);
 
-        } while ! $m2 && 
-                ! $m2->data->{abort} &&
-                defined $state[0]; 
+        } while    ! $m2 
+                && ! $m2->data->{abort} 
+                && defined $state[0]
+                && $redo_count++ < 512
+                ; 
 
         # push capture data
         # print "Concat positional: ", Dumper( $_[3]->data->{match}, $m2->data->{match} );
@@ -155,12 +169,24 @@ sub try_method {
     return eval $sub;
 }
 
+
+sub ignorecase { 
+    my $sub = shift;
+    no warnings qw( uninitialized );
+    return sub {
+        my %param = ( ( defined $_[7] ? %{$_[7]} : () ), ignorecase => 1 );
+        $sub->( @_[0..6], \%param );
+    }
+}
+
 sub constant { 
     my $const = shift;
     my $lconst = length( $const );
     no warnings qw( uninitialized );
     return sub {
-        my $bool = $const eq substr( $_[0], $_[5], $lconst );
+        my $bool = $_[7]{ignorecase}
+            ? lc( $const ) eq lc( substr( $_[0], $_[5], $lconst ) )
+            : $const eq substr( $_[0], $_[5], $lconst );
         $_[3] = Pugs::Runtime::Match->new({ 
                 bool  => \$bool,
                 str   => \$_[0],
@@ -173,10 +199,22 @@ sub constant {
 }
 
 sub perl5 {
-    my $rx = qr(^($_[0]))s;
+    my $rx;
     no warnings qw( uninitialized );
+    { 
+        local $@;
+        $rx = eval " use charnames ':full'; qr(^($_[0]))s ";
+        #print "regex perl5<< $_[0] >>\n";
+        print "Error in perl5 regex: << $_[0] >> \n$@\n"
+            if $@;
+        #die "Error in perl5 regex: $_[0]"
+        #    if $@;
+    }
     return sub {
-        my $bool = substr( $_[0], $_[5] ) =~ m/$rx/;
+        #use charnames ':full';
+        my $bool = $_[7]{ignorecase}
+            ? substr( $_[0], $_[5] ) =~ m/(?i)$rx/
+            : substr( $_[0], $_[5] ) =~ m/$rx/;
         $_[3] = Pugs::Runtime::Match->new({ 
                 bool  => \$bool,
                 str   => \$_[0],
@@ -212,6 +250,7 @@ sub failed {
                 to    => \(0 + $_[5]),
                 named => {},
                 match => [],
+                state => undef,
             });
     }
 };
@@ -406,6 +445,27 @@ sub at_end_of_string {
     }
 };
 
+# experimental!
+sub negate { 
+    my $op = shift;
+    return sub {
+        #my $str = $_[0];
+        my $match = $op->( @_ );
+        my $bool = ! $match;
+
+        $_[3] = Pugs::Runtime::Match->new({ 
+                bool  => \( $bool ),
+                str   => \$_[0],
+                from  => \(0 + $_[5]),
+                to    => \(0 + $_[5]),
+                named => {},
+                match => [],
+                abort => 0,
+            });
+
+    };
+};
+
 # ------- higher-order ruleops
 
 sub optional {
@@ -418,39 +478,140 @@ sub null_or_optional {
     alternation( [ null(), $node ] );
 }
 
+sub greedy_star { 
+    greedy_plus( $_[0], $_[1] || 0, $_[2] ) 
+}
+
+sub non_greedy_star { 
+    non_greedy_plus( $_[0], $_[1] || 0, $_[2] ) 
+}
+
+# XXX - needs optimization for faster backtracking, less stack usage
+# TODO - run-time ranges (iterator)
 sub greedy_plus { 
     my $node = shift;
+    my $min_count = defined( $_[0] ) ? $_[0] : 1;
+    my $max_count = $_[1];  
+    if (  defined $max_count 
+       && $max_count < 1e99
+       ) {
+        return concat( [ 
+            ( $node )             x $min_count, 
+            ( optional( $node ) ) x ($max_count - $min_count) 
+        ] );
+    }
+    # $max_count == infinity
     my $alt;
     $alt = concat( [
         $node, 
         optional( sub{ goto $alt } ),  
     ] );
-    return $alt;
-}
-
-sub greedy_star { 
-    my $node = shift;
-    optional( greedy_plus( $node ) );
-}
-
-sub non_greedy_star { 
-    my $node = shift;
-    alternation( [ 
-        null(),
-        non_greedy_plus( $node ) 
-    ] );
+    return optional( $alt ) if $min_count < 1;
+    return concat( [ ( $node ) x ($min_count - 1), $alt ] );
 }
 
+# XXX - needs optimization for faster backtracking, less stack usage
+# TODO - run-time ranges (iterator)
 sub non_greedy_plus { 
     my $node = shift;
-    # XXX - needs optimization for faster backtracking, less stack usage
+    my $min_count = defined( $_[0] ) ? $_[0] : 1;
+    my $max_count = $_[1] || 1e99;
     return sub {
-        my $state = $_[1] || $node;
-        $state->( $_[0], undef, @_[2..7] );
-        $_[3]->data->{state} = concat( [ $node, $state ] );
+        my $state = $_[1] 
+            || { node  => concat( [ ( $node ) x $min_count ] ), 
+                 count => $min_count };
+        return failed()->(@_)
+            if $state->{count} > $max_count;
+        $state->{node}->( $_[0], undef, @_[2..7] );
+        $_[3]->data->{state} = 
+            { node  => concat( [ $node, $state->{node} ] ), 
+              count => $state->{count} + 1 };
     }
 }
 
+sub range {
+    my $node = shift;
+    my $min_count = shift;
+    my $max_count = shift;
+    my $greedy = not shift;
+    return sub {
+
+        my $continuation = $_[1]; #XXX how do optional continuations work?
+
+        # Forward declarations
+
+        my $try_getting_more;
+
+        my $default_behavior;
+        my $fallback_behavior;
+
+        # Loop variables
+
+        my $count = 0;
+        my $previous_pos = -1;
+
+        # Loop 1 - getting to min_count
+
+        my $continue_towards_min;
+        my $get_minimum = sub {
+            if ( $count < $min_count ) {
+                $count++;
+                goto &$continue_towards_min;
+            } else {
+                goto &$try_getting_more;
+            }
+        };
+        $continue_towards_min = concat( [ $node, $get_minimum ] );
+
+        # Loop 2 - beyond the minimum
+
+        $try_getting_more = sub {
+
+            my $current_pos = $_[5];
+
+            # (1) Stop when max_count is reached, or if pos does not move.
+
+            if ( !( $count < $max_count ) ||
+                 !( $previous_pos < $current_pos ) )
+            {
+                goto &$continuation;
+            }
+            $count++;
+            $previous_pos = $current_pos;
+
+            # (2) Attempt the default behavior.
+
+            # XXX - This section needs to be filled in.
+            # try $default_behavior
+            #  if successful, return.
+            #  if abort, do whatever is needed.
+            #  if fail, we need to backtrack:
+            #    undo any side-effects from trying the $default_behavior,
+            #    so we can do the $fallback_behavior.
+
+            # (3) Since the default behavior failed, do the fall-back beharvior.
+
+            goto &$fallback_behavior;
+
+        };
+        my $get_one_and_maybe_more = concat( [ $node, $try_getting_more ] );
+
+        # Final preparations.
+
+        if ( $greedy ) {
+            $default_behavior = $get_one_and_maybe_more;
+            $fallback_behavior = $continuation;
+        } else { # non-greedy
+            $default_behavior = $continuation;
+            $fallback_behavior = $get_one_and_maybe_more;
+        }
+
+        # Start.
+
+        goto &$get_minimum;
+    };
+}
+
 
 sub preprocess_hash {
     # TODO - move to Pugs::Runtime::Regex
@@ -631,18 +792,6 @@ sub fail {
     );
 };
 
-# experimental!
-sub negate { 
-    my $op = shift;
-    return sub {
-        #my $str = $_[0];
-        my $match = $op->( @_ );
-        return if $match->{bool};
-        return { bool => \1,
-                 #tail => $_[0],
-               }
-    };
-};
 =cut
 
 # experimental!
